use crate::parser;
use proc_macro2::{Ident, TokenStream};
use quote::quote;

use super::{
    api_group::GroupGeneration,
    group::{EnumGeneration, StructGenerator},
    util,
};

#[derive(Debug)]
pub struct GroupMethod<'a> {
    group: &'a GroupGeneration<'a>,
    method: &'a parser::ApiMethod,
}

impl<'a> GroupMethod<'a> {
    pub fn new(group: &'a GroupGeneration, method: &'a parser::ApiMethod) -> Self {
        Self { group, method }
    }

    pub fn generate_method(&self) -> TokenStream {
        let method_name = self.method.name_snake();
        let structs = self.structs();
        let enums = self.enums();
        let builder = self.generate_request_builder();
        let response_struct = self.generate_response_struct();
        let request_method = self.generate_request_method();

        quote! {
            pub mod #method_name {
                #structs
                #enums
                #builder
                #response_struct
                #request_method
            }
        }
    }

    pub fn structs(&self) -> TokenStream {
        let objects = self.method.types.objects();
        let structs = objects
            .iter()
            .map(|obj| StructGenerator::new(obj, self.group).generate_struct());

        quote! {
            #(#structs)*
        }
    }

    pub fn enums(&self) -> TokenStream {
        let enums = self.method.types.enums();
        let generated_enums = enums
            .iter()
            .map(|enum_| EnumGeneration::new(enum_, self.group))
            .map(|e| e.generate());

        quote! {
            #(#generated_enums)*
        }
    }

    fn generate_request_method(&self) -> TokenStream {
        let method_name = self.method.name_snake();

        let parameters = self
            .method
            .types
            .mandatory_params()
            .iter()
            .map(|param| param.to_parameter())
            .collect();

        let form_builder = self.mandatory_parameters_as_form_builder();

        let method_impl = if self.method.types.optional_parameters().is_empty() {
            let builder =
                SendMethodBuilder::new(&method_name, quote! { self.auth }, quote! { form })
                    .set_parameters(parameters)
                    .set_form_factory(quote! {
                        let form = reqwest::multipart::Form::new();
                        #form_builder
                    });

            self.generate_send_method(builder)
        } else {
            quote! {
                pub fn #method_name(&self, #(#parameters),*) -> Builder<'_> {
                    let form = reqwest::multipart::Form::new();
                    #form_builder
                    Builder { group: self, form }
                }
            }
        };

        let group_struct_name = self.group.struct_name();
        let method_impl_with_docs = util::add_docs(&self.method.description, method_impl);

        quote! {
            impl<'a> super::#group_struct_name<'a> {
                #method_impl_with_docs
            }
        }
    }

    fn generate_response_struct(&self) -> TokenStream {
        let response = match self.method.types.response() {
            Some(res) => res,
            None => return quote! {},
        };

        let struct_fields = response
            .types
            .iter()
            .map(|field| field.generate_struct_field());

        let derives = self.group.struct_derives();

        quote! {
            #derives
            pub struct Response {
                #(#struct_fields,)*
            }
        }
    }

    /// Returns a TokenStream containing a request builder if there are optional
    /// parameters, otherwise an empty TokenStream is returned.
    fn generate_request_builder(&self) -> TokenStream {
        let optional_params = self.method.types.optional_parameters();
        if optional_params.is_empty() {
            return quote! {};
        }

        let builder_methods = optional_params
            .iter()
            .map(|param| param.generate_optional_builder_method_with_docs());

        let group_name = self.group.struct_name();
        let send_method = self.generate_send_method(SendMethodBuilder::new(
            &util::to_ident("send"),
            quote! { self.group.auth },
            quote! { self.form },
        ));

        quote! {
            pub struct Builder<'a> {
                group: &'a super::#group_name<'a>,
                form: reqwest::multipart::Form,
            }

            impl<'a> Builder<'a> {
                #send_method
                #(#builder_methods)*
            }
        }
    }

    fn generate_send_method(&self, builder: SendMethodBuilder) -> TokenStream {
        let builder = builder.set_group(self);
        let send_method = builder.generate_send_method();
        let send_method_raw = builder
            .set_override_return_type(quote! { String })
            .append_to_method_name("_raw")
            .generate_send_method();

        let send_method_raw_with_docs = util::add_docs(
            &Some("Returns the raw response of the request.".into()),
            send_method_raw,
        );

        quote! {
            #send_method
            #send_method_raw_with_docs
        }
    }

    fn mandatory_parameters_as_form_builder(&self) -> TokenStream {
        let builder = self
            .method
            .types
            .mandatory_params()
            .into_iter()
            .map(|param| param.generate_form_builder(quote! { form }));

        quote! {
            #(let #builder)*
        }
    }
}

pub struct SendMethodBuilder<'a> {
    group: Option<&'a GroupMethod<'a>>,
    method_name: Ident,
    parameters: Vec<TokenStream>,
    auth_access: TokenStream,
    form_access: TokenStream,
    form_factory: Option<TokenStream>,
    override_return_type: Option<TokenStream>,
}

impl<'a> SendMethodBuilder<'a> {
    pub fn new(method_name: &'a Ident, auth_access: TokenStream, form_access: TokenStream) -> Self {
        Self {
            group: Option::None,
            method_name: method_name.clone(),
            parameters: vec![],
            auth_access,
            form_access,
            form_factory: Option::None,
            override_return_type: Option::None,
        }
    }

    fn generate_send_method(&self) -> TokenStream {
        let method_url = self.method_url();
        let (response_type, response_parse) = self.return_type();
        let form_factory = self.form_factory();
        let parameters = self.parameters.clone();
        let form_access = self.form_access.clone();
        let method_name = self.method_name.clone();
        let auth_access = self.auth_access.clone();

        quote! {
            pub async fn #method_name(self, #(#parameters),*) -> super::super::Result<#response_type> {
                #form_factory
                let res = #auth_access
                    .authenticated_client(#method_url)
                    .multipart(#form_access)
                    .send()
                    .await?
                    #response_parse
                    .await?;

                Ok(res)
            }
        }
    }

    fn return_type(&self) -> (TokenStream, TokenStream) {
        if let Some(override_return_type) = self.override_return_type.clone() {
            (override_return_type, quote! { .text() })
        } else {
            match self.group().method.types.response() {
                Some(resp) => {
                    if resp.is_list {
                        (
                            quote! { std::vec::Vec<Response> },
                            quote! { .json::<std::vec::Vec<Response>>() },
                        )
                    } else {
                        (quote! { Response }, quote! { .json::<Response>() })
                    }
                }
                None => (quote! { String }, quote! { .text() }),
            }
        }
    }

    fn method_url(&self) -> String {
        format!(
            "/api/v2/{}/{}",
            self.group().group.url(),
            self.group().method.url
        )
    }

    fn set_override_return_type(mut self, override_return_type: TokenStream) -> Self {
        self.override_return_type = Some(override_return_type);
        self
    }

    fn set_form_factory(mut self, form_factory: TokenStream) -> Self {
        self.form_factory = Some(form_factory);
        self
    }

    fn form_factory(&self) -> TokenStream {
        self.form_factory.clone().unwrap_or_else(|| quote! {})
    }

    fn set_parameters(mut self, parameters: Vec<TokenStream>) -> Self {
        self.parameters = parameters;
        self
    }

    fn append_to_method_name(mut self, append: &str) -> Self {
        let new_name = util::to_ident(&format!("{}{}", self.method_name, append));
        self.method_name = new_name;
        self
    }

    fn set_group(mut self, group: &'a GroupMethod<'a>) -> Self {
        self.group = Option::Some(group);
        self
    }

    fn group(&self) -> &GroupMethod {
        self.group.expect("Group is not defined")
    }
}
